<?php
/**
 * @file
 * Drupal RestMini module REST client.
 */


class RestMiniClient extends RestMini {

  /**
   * @var integer
   */
  protected static $errorCodeOffset = 1500;

  /**
   * Actual numeric values may be affected by non-zero $errorCodeOffset classes extending RestMini.
   *
   * @see RestMini::$errorCodeOffset
   *
   * @var array $errorCodes
   */
  protected static $errorCodes = array(
    'unknown' => 1,
    'init_connection' => 41,
    'request_options' => 42,
    'parser_not_callable' => 49,
    'response_false' => 51,
    'content_type_mismatch' => 71,
    'response_parse' => 72,
    'response_error' => 81,
  );

  /**
   * Default connect timeout.
   *
   * @var integer
   */
  protected static $connectTimeout = 10;

  /**
   * Default respnse timeout.
   *
   * @var integer
   */
  protected static $responseTimeout = 30;


  /**
   * @var string
   */
  protected $server = '';

  /**
   * @var string
   */
  protected $endpoint = '';

  /**
   * @var array
   */
  protected $options = array();

  /**
   * Last requested url.
   *
   * @var string
   */
  protected $url = '';

  /**
   * @var boolean|NULL
   */
  protected $ssl;

  /**
   * @var array
   */
  protected $responseHeaders = array();

  /**
   *  If error, buckets are:
   *  - (int) code
   *  - (str) name
   *  - (str) message
   *
   * @var array
   */
  protected $error = array();

  /**
   * @var integer
   */
  protected $status = 0;

  /**
   * Response Content-Type header evaluates to NULL if none sent.
   *
   * @var string|NULL
   */
  protected $contentType;

  /**
   * @var string
   */
  protected $parser = 'drupal_json_decode';

  /**
   * Accept request header value.
   *
   * @var string
   */
  protected $parserAccept = 'application/json';

  /**
   * The value returned from parser when failing to parse.
   *
   * @var mixed|NULL
   */
  protected $parserErrorReturn;

  /**
   * @var array
   */
  protected $parserOptions = array();

  /**
   * @var mixed
   */
  protected $response;

  /**
   * @param string $server
   *   Protocol + domain (~ http://ser.ver).
   * @param $endpoint
   *   /path
   * @param array $options
   *   Supported:
   *   - (int) connect_timeout
   *   - (int) response_timeout
   *   - (bool) ssl_verify
   *   - (bool) ignore_status (~ don't trust response status code; like 200 might actually be 404)
   *   - (bool) ignore_content_type (~ don't trust response content type; HTML might actually be JSON, and vice versa)
   *   - (str) user (for [username]:[password])
   *   - (str) pass (for [username]:[password])
   *   - (arr) headers
   *   - (bool) get_headers
   *   - (bool) service_response_info_wrapper (tell service to wrap response in object listing service response properties)
   *   Default: empty.
   */
  public function __construct($server, $endpoint, $options = array()) {
    $this->ssl = strpos($server, 'https') === 0;
    // Remove trailing slash.
    if ($server{strlen($server) - 1} === '/') {
      $server = substr($server, 0, strlen($server) - 1);
    }
    $this->server = $server;
    // Secure leading slash.
    if ($endpoint) {
      $this->endpoint = ($endpoint{0} == '/' ? '' : '/') . $endpoint;
    }
    // Secure timeout options.
    if (!$options) {
      $options = array(
        'connect_timeout' => static::$connectTimeout,
        'response_timeout' => static::$responseTimeout,
      );
    }
    else {
      if (!array_key_exists('connect_timeout', $options)) {
        $options['connect_timeout'] = static::$connectTimeout;
      }
      if (!array_key_exists('response_timeout', $options)) {
        $options['response_timeout'] = static::$responseTimeout;
      }
    }

    $this->options =& $options;
  }

  /**
   * Convenience factory which facilitates chaining.
   *
   * @code
   * // Get JSON-decoded response.
   * $data = RestMiniClient::make('http://server', '/endpoint')->get()->result();
   * // Check status first.
   * $request = RestMiniClient::make('http://server', '/endpoint')->get();
   * if ($request->status() == 200) {
   *   $data = $request->result();
   * }
   * // Get all relevant properties in one go:
   * $response = RestMiniClient::make('http://server', '/endpoint')->get()->result(TRUE);
   * if ($response['status'] == 200) {
   *   // Use $response['result'] ...
   * }
   * elseif (!empty($response['headers']['Some header')) {
   *   // ...
   * }
   * elseif ($response['error']) {
   *   // ...
   * }
   * // Get raw response.
   * $raw = RestMiniClient::make('http://server', '/endpoint')->get()->raw():
   * @endcode
   *
   * @param string $server
   *   Protocol + domain (~ http://ser.ver).
   * @param $endpoint
   *   /path
   * @param array $options
   *   Supported:
   *   - (int) connect_timeout
   *   - (int) response_timeout
   *   - (bool) ssl_verify
   *   - (bool) ignore_status (~ don't trust response status code; like 200 might actually be 404)
   *   - (bool) ignore_content_type (~ don't trust response content type; HTML might actually be JSON, and vice versa)
   *   - (str) user (for [username]:[password])
   *   - (str) pass (for [username]:[password])
   *   - (arr) headers
   *   - (bool) get_headers
   *   - (bool) service_response_info_wrapper (tell service to wrap response in object listing service response properties)
   *   Default: empty.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public static function make($server, $endpoint, $options = array()) {
    return new static($server, $endpoint, $options);
  }

  /**
   * Set and/or remove options.
   *
   * Chainable, returns self.
   *
   * If removing connect_timeout or response_timeout: they will be set again to module defaults.
   *
   * For supported options: see RestMiniClient::make().
   *
   * @code
   * $client = new RestMiniClient('http://ser.ver', '/end-point');
   * // First request, get index.
   * $list = $client->get();
   * // ...
   * // Second request, update a record - but first change some options.
   * $client->alterOptions(
   *   // Set.
   *   array(
   *     'get_headers' => TRUE,
   *   ),
   *   // Remove.
   *   array(
   *     'connect_timeout',
   *     'response_timeout',
   *   ),
   * )->put(
   *   array(21),
   *   NULL,
   *   array('title' => 'Changed record')
   * );
   * @endcode
   *
   * @see RestMiniClient::make()
   *
   * @param array $set
   *   Key-value.
   *   Default: empty.
   * @param array $unset
   *   Non-empty: unset keys named like the values of this array.
   *   Default: empty.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public function alterOptions($set = array(), $unset = array()) {
    foreach ($set as $key => $val) {
      $this->options[$key] = $val;
    }
    foreach ($unset as $val) {
      unset($this->options[$val]);
    }
    // The request method expects these to be present.
    if (!array_key_exists('connect_timeout', $this->options)) {
      $this->options['connect_timeout'] = static::$connectTimeout;
    }
    if (!array_key_exists('response_timeout', $this->options)) {
      $this->options['response_timeout'] = static::$responseTimeout;
    }
    return $this;
  }

  /**
   * Send HTTP GET request, for index or retrieve operation.
   *
   * Chainable, returns self.
   *
   * @see RestMiniClient::make()
   *
   * @param array|NULL $pathParams
   *   Each bucket will be added to the server + endpoint URL.
   *   Example: http://ser.ver/end-point/first-path-param/second-path-param
   *   Default: empty.
   * @param array|NULL $getParams
   *   Each key-value pair becomes key=value.
   *   Example: http://ser.ver/end-point?first=param&second=param
   *   Default: empty.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public function get($pathParams = NULL, $getParams = NULL) {
    $this->request('GET', $pathParams, $getParams);
    return $this;
  }

  /**
   * Send HTTP POST request, for create operation.
   *
   * Chainable, returns self.
   *
   * @see RestMiniClient::make()
   *
   * @param array|NULL $pathParams
   *   Each bucket will be added to the server + endpoint URL.
   *   Example: http://ser.ver/end-point/first-path-param/second-path-param
   *   Default: empty.
   * @param array|NULL $getParams
   *   Each key-value pair becomes key=value.
   *   Example: http://ser.ver/end-point?first=param&second=param
   *   Default: empty.
   * @param array|NULL $postParams
   *   Ignored unless $method is POST or PUT.
   *   Default: empty.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public function post($pathParams = NULL, $getParams = NULL, $postParams = NULL) {
    $this->request('POST', $pathParams, $getParams, $postParams);
    return $this;
  }

  /**
   * Send HTTP PUT request, for update operation.
   *
   * Chainable, returns self.
   *
   * @see RestMiniClient::make()
   *
   * @param array|NULL $pathParams
   *   Each bucket will be added to the server + endpoint URL.
   *   Example: http://ser.ver/end-point/first-path-param/second-path-param
   *   Default: empty.
   * @param array|NULL $getParams
   *   Each key-value pair becomes key=value.
   *   Example: http://ser.ver/end-point?first=param&second=param
   *   Default: empty.
   * @param array|NULL $postParams
   *   Ignored unless $method is POST or PUT.
   *   Default: empty.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public function put($pathParams = NULL, $getParams = NULL, $postParams = NULL) {
    $this->request('PUT', $pathParams, $getParams, $postParams);
    return $this;
  }

  /**
   * Send HTTP DELETE request, for remove or delete operation.
   *
   * Chainable, returns self.
   *
   * @see RestMiniClient::make()
   *
   * @param array|NULL $pathParams
   *   Each bucket will be added to the server + endpoint URL.
   *   Example: http://ser.ver/end-point/first-path-param/second-path-param
   *   Default: empty.
   * @param array|NULL $getParams
   *   Each key-value pair becomes key=value.
   *   Example: http://ser.ver/end-point?first=param&second=param
   *   Default: empty.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public function delete($pathParams = NULL, $getParams = NULL) {
    $this->request('GET', $pathParams, $getParams);
    return $this;
  }

  /*
   * @todo: Support HEAD.
  public function head($pathParams = NULL, $getParams = NULL) {
    $this->request('HEAD', $pathParams, $getParams);
    return $this;
  }
  public function options($pathParams = NULL, $getParams = NULL) {
    $this->request('OPTIONS', $pathParams, $getParams);
    return $this;
  }
  */

  /**
   * Set response parser.
   *
   * Chainable, returns self.
   *
   * @param string $callable
   *   Values: 'function'|'Class::staticMethod'.
   *   Default: drupal_json_decode.
   *
   * @param string $parserAccept
   *   Accept request header value.
   *   Default: application/json.
   *
   * @param mixed $parseErrorReturn
   *   The value returned from parser when failing to parse.
   *   Default: NULL (~ error return value of drupal_json_decode()).
   *
   * @param array $parserOptions
   *   Arguments to pass to the parser, after first argument (the response).
   *   Default: empty.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public function parser($callable = 'drupal_json_decode', $parserAccept = 'application/json', $parseErrorReturn = NULL, $parserOptions = array()) {
    if (!$callable || $callable == 'drupal_json_decode') {
      $this->parser = 'drupal_json_decode';
      $this->parserAccept = 'application/json';
      $this->parserErrorReturn = NULL;
      $this->parserOptions = array();
    }
    else {
      $this->parserOptions = $parserOptions;
      $this->parserAccept = 'application/json';
      $this->parserErrorReturn = $parseErrorReturn;

      $isCallable = TRUE;
      if (($pos = strpos($callable, '::'))) {
        $class = substr($callable, 0, $pos);
        $func = substr($callable, $pos + 2);
        if (!class_exists($class)) {
          $isCallable = FALSE;
          static::log(
            'restmini_client',
            'Parser[' . $callable. '] isnt callable, class[' . $class . '] doesnt exist or isnt included',
            NULL,
            get_object_vars($this),
            WATCHDOG_ERROR
          );
        }
        if (!method_exists($class, $func)) {
          $isCallable = FALSE;
          static::log(
            'restmini_client',
            'Parser[' . $callable. '] isnt callable, method[' . $func . '] doesnt exist',
            NULL,
            get_object_vars($this),
            WATCHDOG_ERROR
          );
        }
      }
      elseif (!function_exists($callable)) {
        $isCallable = FALSE;
        static::log(
          'restmini_client',
          'Parser[' . $callable. '] isnt callable, function[' . $callable . '] doesnt exist',
          NULL,
          get_object_vars($this),
          WATCHDOG_ERROR
        );
      }
      if (!$isCallable) {
        $this->error = array(
          'code' => static::errorCode('parser_not_callable'),
          'name' => 'parser_not_callable',
          'message' => 'Parser ' . $callable . ' isnt callable',
        );
      }
    }

    return $this;
  }

  /**
   * Resets all response properties that may waste memory space.
   *
   * Chainable, returns self.
   *
   * Does not reset parser properties.
   *
   * No need to call this method prior to a new request; the get|post|put|delete() methods call it automatically.
   *
   * @return RestMiniClient
   *   Or extending type.
   */
  public function reset() {
    $this->url = '';
    $this->responseHeaders = array();
    $this->error = array();
    $this->status = 0;
    $this->response = NULL;

    return $this;
  }

  /**
   * Last requested URL.
   *
   * @return string
   *   Empty: no last request, or current request failed before actual sending.
   */
  public function url() {
    return $this->url;
  }

  /**
   * Empty unless request has been sent, and failed.
   *
   *  If error, buckets are:
   *  - (int) code
   *  - (str) name
   *  - (str) message
   *
   * @return array
   */
  public function error() {
    return $this->error;
  }

  /**
   * HTTP response status.
   *
   * @return integer
   *   Zero: request not started, or request failed.
   */
  public function status() {
    return $this->status;
  }

  /**
   * Response headers.
   *
   * @return array()
   *   Empty unless last request used the 'get_headers' option.
   */
  public function headers() {
    return $this->responseHeaders;
  }

  /**
   * Get raw response.
   *
   * @return string|boolean|NULL
   *   NULL: request not started.
   *   FALSE: request failed.
   */
  public function raw() {
    return $this->response;
  }

  /**
   * Parsed response, or array of most properties of the response.
   *
   * If $fetchKeyPath, and it doesn't match: returns the whole parsed response.
   *
   *  Returned wrapper array buckets when $responseInfo:
   *  - (int) status
   *  - (str|null) content_type
   *  - (arr) headers
   *  - (mixed) result
   *  - (arr) error
   *
   * @code
   * $get_result_key = RestMiniClient::make('http://server', '/endpoint')->get()->result(array('remote', 'server', 'wraps', 'my', 'data'));
   * @endcode
   *
   * @param array $fetchKeyPath
   *   List of keys to recurse by to find the actual payload data.
   *   Default: empty.
   * @param boolean $responseInfo
   *   Ignored if $fetchKeyPath, unless error (then response info may be useful).
   *   Truthy: get all properties of the response.
   *   Default: FALSE (~ get result only).
   *
   * @return mixed
   *   NULL: request not started, or failed to JSON-decode response.
   *   FALSE: request failed, or actual parsed response.
   *   Empty string: empty response.
   */
  public function result($fetchKeyPath = array(), $responseInfo = FALSE) {
    if ($this->error) {
      return !$responseInfo ? FALSE : array(
        'status' => $this->status,
        'content_type' => $this->contentType,
        'headers' => $this->responseHeaders,
        'url' => $this->url,
        'result' => $this->response,
        'error' => $this->error,
      );
    }

    // Empty.
    if ($this->response == '') {
      return !$responseInfo ? '' : array(
        'status' => $this->status,
        'content_type' => $this->contentType,
        'headers' => $this->responseHeaders,
        'url' => $this->url,
        'result' => '',
        'error' => array(),
      );
    }

    // Get out if status indicates no usable content.
    if (!$responseInfo && $this->status >= 300 && empty($this->options['ignore_status'])) {
      return '';
    }

    // Detect HTML if we don't want that - lots of services return HTML as fallback when erring.
    $parse = TRUE;
    if (!empty($this->options['ignore_content_type'])) {
      if (!strpos($this->parserAccept, 'xml') && $this->parserAccept != 'text/html' && $this->response{0} === '<') {
        $parse = FALSE;
      }
    }
    elseif ($this->contentType != $this->parserAccept) {
      $parse = FALSE;
    }
    if (!$parse) {
      $this->error = array(
        'code' => static::errorCode('content_type_mismatch'),
        'name' => 'content_type_mismatch',
        'message' => 'Response content type doesnt match parser',
      );
      return !$responseInfo ? '' : array(
        'status' => $this->status,
        'content_type' => $this->contentType,
        'headers' => $this->responseHeaders,
        'url' => $this->url,
        'result' => $this->response,
        'error' => $this->error,
      );
    }

    // Parse.
    $parser = $this->parser;
    if (!strpos($parser, ':') && !$this->parserOptions) {
      $data = $parser($this->response);
    }
    else {
      if (!$this->parserOptions) {
        $data = call_user_func($parser, $this->response);
      }
      else {
        $args = $this->parserOptions; // Copy.
        array_unshift($args, $this->response);
        $data = call_user_func_array($parser, $args);
      }
    }

    // Parse error.
    if ($data === $this->parserErrorReturn) {
      static::log(
        'restmini_client',
        'Failed to parse response',
        NULL,
        get_object_vars($this),
        WATCHDOG_ERROR
      );
      $this->error = array(
        'code' => static::errorCode('response_parse'),
        'name' => 'response_parse',
        'message' => 'Failed to parse response',
      );
      return !$responseInfo ? FALSE : array(
        'status' => $this->status,
        'content_type' => $this->contentType,
        'headers' => $this->responseHeaders,
        'url' => $this->url,
        'result' => FALSE,
        'error' => $this->error,
      );
    }

    if (!$fetchKeyPath) {
      return !$responseInfo ? $data : array(
        'status' => $this->status,
        'content_type' => $this->contentType,
        'headers' => $this->responseHeaders,
        'url' => $this->url,
        'result' => $data,
        'error' => array(),
      );
    }

    // Copy the whole data set in case the keypath doesn't match.
    $orig = $data;
    // Recurse.
    foreach ($fetchKeyPath as $key) {
      if ($data && is_array($data) && array_key_exists($key, $data)) {
        $data = $data[$key];
      }
      else {
        // No match, return all.
        return $orig;
      }
    }
    return $data;
  }


  /**
   * Adjust endpoint before sending request.
   *
   * Mustn't change instance var endpoint, must only return that or an adjusted version of it.
   *
   * Meant to overridden in extending class.
   *
   * @see RestMiniClient::request()
   *
   * @return string
   */
  protected function endpointAdjust() {
    return $this->endpoint;
  }

  /**
   * Resets instance vars that may get populated upon request.
   *
   * @throws Exception
   *
   * @param $method
   *   GET ~ index|retrieve (default).
   *   POST ~ create.
   *   PUT ~ update.
   *   DELETE.
   * @param array|NULL $pathParams
   *   Default: empty.
   * @param array|NULL $getParams
   *   Default: empty.
   * @param array|NULL $postParams
   *   Ignored unless $method is POST or PUT.
   *   Default: empty.
   *
   * @return boolean
   *   False on error; logs error and populates instance var error.
   */
  protected function request($method = 'GET', $pathParams = NULL, $getParams = NULL, $postParams = NULL) {
    // Reset instance.
    $this->reset();

    $this->url = $this->server . $this->endpointAdjust();

    // Path parameters will get double url encoding of some chars, to prevent parsing errors;
    // a slash in a path parameter could otherwise be interpreted as two path fragments instead of one
    // (Drupal url-decodes full url before parsing into path fragments and get parameters).
    if ($pathParams) {
      foreach ($pathParams as $val) {
        $this->url .= '/' . rawurlencode(str_replace(array('/', '?', '&', '='), array('%2F', '3F', '26', '3D'), $val));
      }
    }

    // GET parameters get single encoding.
    if ($getParams) {
      $i = -1;
      foreach ($getParams as $key => $val) {
        // Only ?-delimiter if first param and (empty endpoint or endpoint doesn't contain ?).
        $this->url .= (!(++$i) && (!$this->endpoint || !strpos($this->endpoint, '?')) ? '?' : '&')
          . $key . '=' . rawurlencode($val);
      }
    }

    // Options.
    $options =& $this->options;
    $curlOpts = array(
      // Don't include header in output.
      CURLOPT_HEADER => FALSE,
      // Get response as string, don't echo it.
      CURLOPT_RETURNTRANSFER => TRUE,
      // Follow redirects.
      CURLOPT_FOLLOWLOCATION => TRUE,
      // Timeouts.
      CURLOPT_CONNECTTIMEOUT => $options['connect_timeout'],
      CURLOPT_TIMEOUT => $options['response_timeout'],
    );

    // Getting response header comes with a performance price tag, so we only do it on demand.
    if (!empty($options['get_headers'])) {
      $curlOpts[CURLOPT_HEADERFUNCTION] = array($this, 'responseHeaderCallback');
    }
    // Don't verify SSL certificate?
    if ($this->ssl && array_key_exists('ssl_verify', $options) && !$options['ssl_verify']) {
      $curlOpts[CURLOPT_SSL_VERIFYPEER] = FALSE;
    }
    // user:pass?
    if (!empty($options['user']) && !empty($options['pass'])) {
      $curlOpts[CURLOPT_USERPWD] = $options['user'] . ':' . $options['pass'];
    }
    // Headers.
    if (empty($options['headers'])) {
      $headers = array(
        'Accept: ' . $this->parserAccept
      );
    }
    else {
      // Copy, because we may add header(s), per request.
      $headers = $options['headers'];
      $headersConcat = array();
      foreach ($headers as $key => $val) {
        $headersConcat[] = $key . ': ' . $val;
      }
      if (empty($headers['Accept']) && empty($headers['accept'])) {
        $headersConcat[] = 'Accept: ' . $this->parserAccept;
      }
      $headers =& $headersConcat;
    }
    if (!empty($options['service_response_info_wrapper'])) {
      $headers[] = 'X-Rest-Service-Response-Info-Wrapper: 1';
    }
    $curlOpts[CURLOPT_HTTPHEADER] =& $headers;

    // http method.
    switch ($method) {
      case 'POST':
        $curlOpts[CURLOPT_POST] = TRUE;
        if ($postParams) {
          $curlOpts[CURLOPT_POSTFIELDS] = $postParams;
        }
        break;
      case 'PUT':
        $curlOpts[CURLOPT_PUT] = TRUE;
        if ($postParams) {
          $curlOpts[CURLOPT_POSTFIELDS] = $postParams;
        }
        break;
      case 'DELETE':
        $curlOpts[CURLOPT_CUSTOMREQUEST] = 'DELETE';
        break;
      default:
        $method = 'GET';
    }

    // cUrl begin.
    $resource = curl_init($this->url);
    if ($resource === FALSE) {
      static::log(
        'restmini_client',
        'Failed to initiate connection',
        NULL,
        array(
          'server' => $this->server,
          'endpoint' => $this->endpoint,
          'options' => $options,
          'method' => $method,
          'path params' => $pathParams,
          'GET params' => $getParams,
          'POST params' => $postParams,
          'url' => $this->url,
        ),
        WATCHDOG_ERROR
      );
      $this->error = array(
        'code' => static::errorCode('init_connection'),
        'name' => 'init_connection',
        'message' => 'Failed to initiate connection',
      );

      return FALSE;
    }

    // Set options.
    if (!curl_setopt_array($resource, $curlOpts)) {
      static::log(
        'restmini_client',
        'Failed to set request options',
        NULL,
        array(
          'server' => $this->server,
          'endpoint' => $this->endpoint,
          'options' => $options,
          'method' => $method,
          'path params' => $pathParams,
          'GET params' => $getParams,
          'POST params' => $postParams,
          'url' => $this->url,
          'curl info' => curl_getinfo($resource),
        ),
        WATCHDOG_ERROR
      );
      curl_close($resource);
      $this->error = array(
        'code' => static::errorCode('request_options'),
        'name' => 'request_options',
        'message' => 'Failed to set request options',
      );

      return FALSE;
    }
    unset($curlOpts);

    // Send request.
    $this->response = curl_exec($resource);

    // Get status code.
    $this->status = curl_getinfo($resource, CURLINFO_HTTP_CODE);
    // Content type may be NULL (none), and it may contain character set (~ text/html; charset=utf-8).
    if (($contentType = curl_getinfo($resource, CURLINFO_CONTENT_TYPE)) && ($pos = strpos($contentType, ';'))) {
      $contentType = substr($contentType, 0, $pos);
    }
    $this->contentType = $contentType;

    if ($this->response === FALSE) {
      static::log(
        'restmini_client',
        'Response is false'
          . ((!$this->ssl || (array_key_exists('ssl_verify', $options) && !$options['ssl_verify'])) ? '' : ', check ssl_verify'),
        NULL,
        array(
          'status' => $this->status,
          'response headers' => $this->responseHeaders,
          'server' => $this->server,
          'endpoint' => $this->endpoint,
          'options' => $options,
          'method' => $method,
          'path params' => $pathParams,
          'GET params' => $getParams,
          'POST params' => $postParams,
          'url' => $this->url,
          'curl info' => curl_getinfo($resource),
        ),
        WATCHDOG_ERROR
      );
      curl_close($resource);
      $this->error = array(
        'code' => static::errorCode('response_false'),
        'name' => 'response_false',
        'message' => 'Response is false',
      );

      return FALSE;
    }
    // ...else: Response must be string, because of CURLOPT_RETURNTRANSFER.
    $this->response = trim($this->response);

    curl_close($resource);

    // Check for error status.
    if ($this->status >= 500) {
      static::log(
        'restmini_client',
        'Response error status code ' . $this->status,
        NULL,
        array(
          'status' => $this->status,
          'response headers' => $this->responseHeaders,
          'server' => $this->server,
          'endpoint' => $this->endpoint,
          'options' => $options,
          'method' => $method,
          'path params' => $pathParams,
          'GET params' => $getParams,
          'POST params' => $postParams,
          'url' => $this->url,
        ),
        WATCHDOG_ERROR
      );
      $this->error = array(
        'code' => static::errorCode('response_error'),
        'name' => 'response_error',
        'message' => 'Response error',
      );
      return FALSE;
    }

    return TRUE;
  }

  /**
   * CURLOPT_HEADERFUNCTION implementation.
   *
   * @param resource $resource
   * @param string $headerLine
   * @return integer
   *   Header line byte length.
   */
  protected function responseHeaderCallback($resource, $headerLine) {
    // Remove trailing \r\n.
    if (($line = trim($headerLine))) {
      if (!isset($this->responseHeaders['status_text']) && strpos($line, 'HTTP') === 0) {
        if (strpos($line, ' OK') == strlen($line) - 3) {
          $this->responseHeaders['status_text'] = 'OK';
        }
        else {
          $this->responseHeaders['status_text'] = preg_replace('/^HTTPS?\/\d\.\d+ \d+ (.+)$/', '$1', $line);
        }
      }
      elseif (($pos = strpos($line, ': '))) {
        $this->responseHeaders[substr($line, 0, $pos)] = substr($line, $pos + 2);
      }
      else {
        $this->responseHeaders[] = $line;
      }
    }
    // Satisfy return value contract with PHP cUrl.
    return strlen($headerLine);
  }

}
